#include <cstdio>
#include <cstring>
#include <streambuf>
using namespace std;

extern "C" {
    int read (int fd, char* buf, int anz);
}

class inbuf : public streambuf {
  protected:
    /* Datenpuffer:
     *  maximal 4 Zeichen Putback-Bereich plus
     *  maximal 6 Zeichen normaler Lesepuffer
     */
    const int pufferSize = 10;     // Groesse des Datenpuffers
    char puffer[pufferSize];       // Datenpuffer

  public:
    /* Konstruktor
     *  - Datenpuffer leer initialisieren
     *  - keine Putback-Reserve
     *  => underflow() forcieren
     */
    inbuf() {
        setg (puffer+4,     // Putback-Anfang
              puffer+4,     // Leseposition
              puffer+4);    // Puffer-Ende
    }

  protected:
    /* neue Zeichen in den Puffer einlesen
     */
    virtual int_type underflow () {

        // Leseposition vor Puffer-Ende ?
        if (gptr() < egptr()) {
            return *gptr();
        }

        /* Anzahl Putback-Bereich ermitteln
         *  - Anzahl bereits gelesener Zeichen
         *  - maximal 4
         */
        int anzPutback;
        anzPutback = gptr() - eback();
        if (anzPutback > 4) {
            anzPutback = 4;
        }

        /* die bis zu vier vorher gelesenen Zeichen nach vorne
         * in den Putback-Puffer (erste 4 Zeichen)
         */
        memcpy (puffer+(4-anzPutback), gptr()-anzPutback, anzPutback);

        /* neue Zeichen lesen
         */
        int anz;
        anz = read (0, puffer+4, pufferSize-4);
        if (anz <= 0) {
            // Fehler oder EOF
            return EOF;
        }

        /* Puffer-Zeiger neu setzen
         */
        setg (puffer+(4-anzPutback),   // Putback-Anfang
              puffer+4,                // Leseposition
              puffer+4+anz);           // Puffer-Ende

        // naechstes Zeichen zurueckliefern
        return *gptr();
    }
};
